
_q_exports = ["simple", "error"]

import sys
from quixote.demo.pages import _q_index
from quixote.demo.integer_ui import IntegerUI

def simple (request):
    # This function returns a plain text document, not HTML.
    request.response.set_header("Content-type", "text/plain")
    return "This is the Python function 'quixote.demo.simple'.\n"

def error (request):
    raise ValueError, "this is a Python exception"

def _q_getname(request, component):
    return IntegerUI(request, component)
