#!/usr/bin/env python

# Setup script for Quixote

__revision__ = "$Id: setup.py,v 1.12 2001/10/10 18:01:41 gward Exp $"

from distutils.core import setup
from qx_distutils import qx_build_py

setup (name = "Quixote",
       version = "0.4.1",
       description = "A highly Pythonic Web application framework",
       author = "MEMS Exchange",
       author_email = "quixote@mems-exchange.org",
       url = "http://www.mems-exchange.org/software/python/quixote/",

       package_dir = {'quixote':''},
       packages = ['quixote',  'quixote.demo', 'quixote.form'],

       cmdclass = {'build_py': qx_build_py},
      )

 
